/**
 * The synchronizer-ws module of the TinyBase project lets you synchronize
 * MergeableStore data to and from other MergeableStore instances via WebSockets
 * facilitated by a server.
 * @see Synchronizing Data guide
 * @see Todo App v6 (collaboration) demo
 * @packageDocumentation
 * @module synchronizer-ws-client
 * @since v5.0.0
 */

import {MergeableStore} from '../mergeable-store.d.ts';
import {Synchronizer} from '../synchronizers.d.ts';
import {WebSocket as WsWebSocket} from 'ws';

export type WebSocketTypes = WebSocket | WsWebSocket;

/**
 * The WsSynchronizer interface
 * @category Synchronizer
 */
export interface WsSynchronizer<WebSocketType extends WebSocketTypes>
  extends Synchronizer {
  getWebSocket(): WebSocketType;
}

/**
 * The createWsSynchronizer function
 * @category Creation
 */
export function createWsSynchronizer<WebSocketType extends WebSocketTypes>(
  store: MergeableStore,
  webSocket: WebSocketType,
  requestTimeoutSeconds?: number,
  onIgnoredError?: (error: any) => void,
): Promise<WsSynchronizer<WebSocketType>>;
