/**
 * The synchronizer-ws-server module of the TinyBase project lets you create
 * a server that facilitates synchronization between clients.
 * @see Synchronizing Data guide
 * @see Todo App v6 (collaboration) demo
 * @packageDocumentation
 * @module synchronizer-ws-server
 * @since v5.0.0
 */

import {Id, IdOrNull, Ids} from '../common.d.ts';
import {GetIdChanges} from '../store.d.ts';
import {WebSocketServer} from 'ws';

/**
 * The PathIdsListener type
 */
export type PathIdsListener<> = (
  wsServer: WsServer,
  getIdChanges: GetIdChanges,
) => void;

/**
 * The ClientIdsListener type
 */
export type ClientIdsListener<> = (
  wsServer: WsServer,
  pathId: Id,
  getIdChanges: GetIdChanges,
) => void;

/**
 * The WsServerStats type
 */
export type WsServerStats = {
  paths?: number;
  clients?: number;
};

/**
 * The WsServer interface
 */
export interface WsServer {
  getWebSocketServer: () => WebSocketServer;
  getPathIds: () => Ids;
  getClientIds: (pathId: Id) => Ids;
  addPathIdsListener: (listener: PathIdsListener) => Id;
  addClientIdsListener: (pathId: IdOrNull, listener: ClientIdsListener) => Id;
  delListener(listenerId: Id): WsServer;
  getStats: () => WsServerStats;
  destroy: () => void;
}

/**
 * The createWsServer function
 * @category Creation
 */
export function createWsServer(WebSocketServer: WebSocketServer): WsServer;
